#include <avr/io.h>
#include <util/delay.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include "nokia5110.h"

#define DEFAULT_DELAY 200

#define LEFT PD5
#define RIGHT PD6
#define UP PD0
#define DOWN PD4

#define CAR_RADIUS 5
#define LANE_1 32	// 	LANE_1 = player.x - CAR_RADIUS * 2;
#define LANE_2 42	//	LANE_2 = player.x;
#define LANE_3 52	//	LANE_3 = player.x + CAR_RADIUS * 2;
#define MAX_ENEMIES 4

#define MAX_SCREEN_X 84
#define MAX_SCREEN_Y 48
typedef struct car {
	uint8_t y;
	uint8_t lane;
} Car;

typedef struct car_enemy {
	uint8_t y;
	uint8_t lane;
} Car_enemy;

static unsigned int time = 0;

void init_buttons() {

	DDRD &= ~(1 << UP);
	PORTD |= (1 << UP);

	DDRD &= ~(1 << DOWN);
	PORTD |= (1 << DOWN);
	
	DDRD &= ~(1 << LEFT);
	PORTD |= (1 << LEFT);

	DDRD &= ~(1 << RIGHT);
	PORTD |= (1 << RIGHT);

}

uint8_t check_pressed_up_button() {
	 if ((PIND & (1 << UP)) == 0) 
	 	return 1;
	 return 0;
}

uint8_t check_pressed_down_button() {
	 if ((PIND & (1 << DOWN)) == 0) 
	 	return 1;
	 return 0;
}

uint8_t check_pressed_left_button() {
	 if ((PIND & (1 << LEFT)) == 0) {
	 	return 1;
	 }
	 return 0;
}

uint8_t check_pressed_right_button() {
	 if ((PIND & (1 << RIGHT)) == 0) 
	 	return 1;
	 return 0;
}

uint8_t generateNr(const int nMin, const int nMax) {
 	int nrRandomNumber;
 	srand(time);
 	time ++;
 	if (time > 5000)
 		time = 0;
	nrRandomNumber = rand()%(nMax-nMin) + nMin;
    return nrRandomNumber;
}

void draw_player (Car player) {
	uint8_t i,j;
	uint8_t x1,x2,y1,y2;
	x1 = player.lane - CAR_RADIUS + 1;
	x2 = player.lane + CAR_RADIUS - 1;
	y1 = player.y - CAR_RADIUS;
	y2 = player.y + CAR_RADIUS;
	for (i = x1; i <= x2; i++)
		for (j = y1; j <= y2; j++) 
			if (i != x1 + 2 && i != x2 - 2)
					nokia_lcd_set_pixel(i,j,1);	

	for (j = y1 + 2; j <= y2 - 2; j++) {
		nokia_lcd_set_pixel(x1,j,0);
		nokia_lcd_set_pixel(x2,j,0);
		nokia_lcd_set_pixel(x1 + 1,j,0);
		nokia_lcd_set_pixel(x2 - 1,j,0);
	}

}
	
void draw_enemy (Car_enemy enemy) {
	uint8_t i,j;
	uint8_t x1,x2,y1,y2;

	x1 = enemy.lane - CAR_RADIUS + 1;
	x2 = enemy.lane + CAR_RADIUS - 1;

	y1 = enemy.y - CAR_RADIUS;
	y2 = enemy.y + CAR_RADIUS;

	for (i = x1; i <= x2; i++)
		for (j = y1; j <= y2; j++) 
			if (i != x1 + 2 && i != x2 - 2)
					nokia_lcd_set_pixel(i,j,1);	

	for (j = y1 + 2; j <= y2 - 2; j++) {
		nokia_lcd_set_pixel(x1,j,0);
		nokia_lcd_set_pixel(x2,j,0);
		nokia_lcd_set_pixel(x1 + 1,j,0);
		nokia_lcd_set_pixel(x2 - 1,j,0);
	}

}

void delay_ms(unsigned int milliseconds)
 {
   while(milliseconds > 0)
   {
    	_delay_ms(1);
      	milliseconds--;
    }
 }

void start_game() {
	nokia_lcd_set_cursor(5,5);
	nokia_lcd_write_string("Press UP key",1);
	nokia_lcd_set_cursor(20,15);
	nokia_lcd_write_string("to start",1);
	nokia_lcd_set_cursor(15,25);
	nokia_lcd_write_string("CAR RACING",1);
	nokia_lcd_set_cursor(30,35);
	nokia_lcd_write_string("GAME",1);
	nokia_lcd_render();
	delay_ms(10);
	while(1) {
	if (check_pressed_up_button() == 1)
			break;
	else{
		time++;
		delay_ms(20);
	}
	}	
}

void game_over() {
	nokia_lcd_set_cursor(10,10);
	nokia_lcd_write_string("GAME",2);
	nokia_lcd_set_cursor(30,30);
	nokia_lcd_write_string("OVER",2);
	nokia_lcd_render();
}

uint8_t collision_detector(Car_enemy enemies[MAX_ENEMIES],uint8_t enemies_active[MAX_ENEMIES],Car player) {
	uint8_t i;
	for (i = 0; i < MAX_ENEMIES;i++)
		if (enemies_active[i] == 1)
			if (player.lane == enemies[i].lane && ((player.y - CAR_RADIUS / 2) - (enemies[i].y + CAR_RADIUS / 2)) < 5 )
				return 1;
	return 0;
}

int main(void)
{
	uint8_t i,j;
	uint8_t current_lane;
	uint8_t pressed_button = 0;
	uint8_t nr_current_enemies = 0;
	uint8_t t = 0;
	uint8_t generated_result = 0;
	uint8_t collision_flag = 0;
	uint8_t delay_time = DEFAULT_DELAY;
	uint8_t last_enemies_lanes[3], k = 0;
    init_buttons();
	nokia_lcd_init();
	nokia_lcd_power(1);
	nokia_lcd_clear();
	Car player;
	Car_enemy enemies[MAX_ENEMIES];
	uint8_t enemies_active[MAX_ENEMIES]; // 0 inactive; 1 active
	for (i = 0; i < MAX_ENEMIES;i++) {
		enemies[i].y = CAR_RADIUS;
		enemies_active[i] = 0;
	}

	player.lane = LANE_2;
	player.y = MAX_SCREEN_Y - CAR_RADIUS - 2;
	
	last_enemies_lanes[0] = 0;
	last_enemies_lanes[1] = 0;
	last_enemies_lanes[2] = 0;
	
	start_game();
	while (1) {

		pressed_button = 0;
		nokia_lcd_clear();
		delay_ms(10);

		if (check_pressed_left_button() == 1) {
			if (player.lane == LANE_1) {
				;
			}
			else if (player.lane == LANE_2) {
				player.lane = LANE_1;
			}
			else if (player.lane == LANE_3) {
				player.lane = LANE_2;
			}
			pressed_button = 1;
		}

		if (check_pressed_right_button() == 1) {
			if (player.lane == LANE_1)
				player.lane = LANE_2;
		
			else if (player.lane == LANE_2)
				player.lane = LANE_3;
			
			else if (player.lane == LANE_3)
				;
	
			pressed_button = 1;
		}

		if (check_pressed_up_button() == 1) {
			if(delay_time >= 20)
				delay_time -= 10;
		}

		if (check_pressed_down_button() == 1) {
			if(delay_time <= DEFAULT_DELAY)
				delay_time += 10;
		}
		if (pressed_button == 1)
			delay_ms(10);

		t++;

		if (nr_current_enemies == 0 || (t % 10 == 0 && nr_current_enemies < MAX_ENEMIES)) {
			for (i = 0; i < MAX_ENEMIES; i++)
				if (enemies_active[i] == 0) {
					generated_result = generateNr(0,6);
					if (generated_result <= 1) {
						enemies[i].lane = LANE_1;
						last_enemies_lanes[k] = 1;
						k++;
					}
					else if (generated_result == 2 || generated_result == 3) {
						enemies[i].lane = LANE_2;
						last_enemies_lanes[k] = 2;
						k++;
					}
					else if (generated_result >= 4) {
						enemies[i].lane = LANE_3;
						last_enemies_lanes[k] = 3;
						k++;
					}
					if (last_enemies_lanes[0] == last_enemies_lanes[1] && last_enemies_lanes[1] == last_enemies_lanes[2]) {
						i -= 2;
						k--;
						continue;
					}
					enemies[i].y = CAR_RADIUS;
					enemies_active[i] = 1;
					nr_current_enemies++;
					break;
		}
	}
	for (i = 0; i < nr_current_enemies; i++) {
		if (enemies_active[i] == 1) {
			if (t % 2 == 0)
				enemies[i].y += CAR_RADIUS;
			draw_enemy(enemies[i]);
			if (enemies[i].y >= MAX_SCREEN_Y) {
				enemies[i].y = CAR_RADIUS;
				enemies_active[i] = 0;
				nr_current_enemies--;
			}
		}

	}

	draw_player(player);
		for (i = 0; i < MAX_SCREEN_X; i++)
			for (j = 0; j < MAX_SCREEN_Y;j++)
			if (((i + j) % 2 == 0) && (i < LANE_1 - CAR_RADIUS - 4 || i > LANE_3 + CAR_RADIUS + 4))
				nokia_lcd_set_pixel(i,j,1);

			if(i == LANE_1 - CAR_RADIUS - 1 || i == LANE_3 + CAR_RADIUS + 1 )
				nokia_lcd_set_pixel(i,j,1);
			if(i == LANE_1 - CAR_RADIUS - 3 || i == LANE_3 + CAR_RADIUS + 3 )
				nokia_lcd_set_pixel(i,j,0);

	delay_ms(delay_time);
	nokia_lcd_render();
	if (t > 5000)
		t = 0;
	collision_flag = collision_detector(enemies,enemies_active,player);
	if (collision_flag != 0) {
		game_over();
		_delay_ms(9999);
		return 0;
		}
	}
}
    